/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.util;

import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;
import org.springframework.stereotype.Component;

import java.io.IOException;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Properties;

@Component
public class RoleMappingProperties {
    private static final Logger log = LogManager.getLogger(RoleMappingProperties.class);

    private static String ROLE_MAPPING_FILE_NAME = "/nsp-role-mapping.properties";
    private static final String CITIZEN_USER = "citizen.user";
    private static final String HEALTHCAREPROFESSIONAL_UNKNOWN_USER = "healthcareprofessional.unknown.user";

    private static final String ERROR_READING_FILE = "Error reading role mapping configuration properties from: %s";
    private static final String ERROR_READING_PROPERTY = "Error reading role mapping configuration from one of the properties %s from: %s";


    private Map<String, String> mappingRoles = new LinkedHashMap<>();

    RoleMappingProperties(String fileName) { //package scope to allow unit testing of error scenario
        loadRoleMappingProperties(fileName);
    }
    public RoleMappingProperties() {
        loadRoleMappingProperties(ROLE_MAPPING_FILE_NAME);
    }
    public void loadRoleMappingProperties(String fileName) {
        java.util.Properties properties = new Properties();
        try {
            log.debug("Loading role mapping configuration from " + fileName);
            properties.load(RoleMappingProperties.class.getResourceAsStream(fileName));
            for (String key : properties.stringPropertyNames()) {
                mappingRoles.put(key, properties.getProperty(key));
                log.debug("Role-key: " + key + " Role-value: " + properties.getProperty(key));
            }

        } catch (IOException e) {
            throw new RuntimeException(String.format(ERROR_READING_FILE, fileName), e);
        }
        if (getBorgerUser() == null || getUnknownUser() == null) {
            throw new RuntimeException(String.format(ERROR_READING_PROPERTY, CITIZEN_USER + ", " + HEALTHCAREPROFESSIONAL_UNKNOWN_USER, fileName));
        }
    }

    public Map<String, String> getAllMappingValues() {
        return mappingRoles;
    }

    public String getBorgerUser() {
        return mappingRoles.get(CITIZEN_USER);
    }

    public String getUnknownUser() {
        return mappingRoles.get(HEALTHCAREPROFESSIONAL_UNKNOWN_USER);
    }

}
