/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.context;

import dk.sds.nsp.security.DevelopmentSecurityProvider;
import dk.sds.nsp.security.SecurityContext;
import dk.sundhedsdatastyrelsen.livstestamente.domain.UserType;
import dk.sundhedsdatastyrelsen.livstestamente.util.FaultCode;
import dk.sundhedsdatastyrelsen.livstestamente.util.RoleMappingProperties;
import org.junit.Before;
import org.junit.Test;
import org.mockito.Mockito;

import java.util.*;

import static org.junit.Assert.*;
import static org.junit.Assert.assertNull;

public class ServiceActorProviderTest {

    private SecurityContext securityContext;
    private RoleMappingProperties roleMappingProperties;
    private UserTitleProvider userTitleProvider;
    private ServiceActorProvider serviceActorProvider;
    private List<String> whitelistedCvrs;
    private String audience;

    private final String BORGER_USER_TITLE = "borger";
    private final String UNKNOWN_USER_TITLE = "unverified role";
    private Map<String, String> MAPPING_ROLES = new LinkedHashMap<String, String>(){{
        put("5151", "Fysioterapeut");
        put("5265", "Kiropraktor");
        put("urn:dk:healthcare:national-federation-role:code:41001:value:SundAssistR1", "Sundhedsstamkort-Læser");
        put("urn:dk:healthcare:national-federation-role:code:41002:value:SundAssistR2", "Sundhedsdata-Læser");
        put("citizen.user.title", "Borger");
        put("healthcareprofessional.unknown.user.title", "unverified role");
    }};

    @Before
    public void setUp() throws Exception {
        setUpSecurityContext();

        audience = securityContext.getTicket().get().getAudience().get();
        whitelistedCvrs = Collections.singletonList(securityContext.getOrganisation().get().getIdentifier());

        roleMappingProperties = Mockito.mock(RoleMappingProperties.class);
        Mockito.when(roleMappingProperties.getBorgerUser()).thenReturn(BORGER_USER_TITLE);
        Mockito.when(roleMappingProperties.getUnknownUser()).thenReturn(UNKNOWN_USER_TITLE);
        Mockito.when(roleMappingProperties.getAllMappingValues()).thenReturn(MAPPING_ROLES);

        userTitleProvider = new UserTitleProvider(roleMappingProperties);
        serviceActorProvider = new ServiceActorProvider(audience, whitelistedCvrs, userTitleProvider);
    }

    private void setUpSecurityContext() {
        securityContext = Mockito.spy(new DevelopmentSecurityProvider());
        Mockito.when(securityContext.getPrincipalUser()).thenReturn(Optional.empty());
    }

    @Test
    public void getServiceActor() {
        // Act
        ServiceActor serviceActor = serviceActorProvider.getServiceActor(securityContext, false);

        // Assert
        SecurityContext.User actingUser = securityContext.getActingUser().get();
        SecurityContext.Organisation organisation = securityContext.getOrganisation().get();
        assertEquals(actingUser.getIdentifier(), serviceActor.getPersonIdentifier());
        assertEquals(actingUser.getGivenName().get(), serviceActor.getFirstName());
        assertEquals(actingUser.getSurname().get(), serviceActor.getLastName());
        assertEquals(actingUser.getCredentials().get().getAuthorizationCode().get(), serviceActor.getAuthorisationNumber());
        assertEquals(actingUser.getCredentials().get().getEducationCode().get(), serviceActor.getEducationCode());
        assertEquals(organisation.getIdentifier(), serviceActor.getOrganizationId());
        assertEquals("CVR", serviceActor.getOrganizationIdSource().toString());
        assertEquals(organisation.getName().get(), serviceActor.getOrganizationName());
        assertEquals(UserType.HealthCareProfessional, serviceActor.getUserType());
        assertEquals("7170", serviceActor.getUserTitle());
    }

    @Test
    public void getServiceActor_forHealthCareProfessionalWithEducationCode() {
        // Act
        mockActingHealthcareProfessionalUser("1234", null);
        ServiceActor serviceActor = serviceActorProvider.getServiceActor(securityContext, false);

        // Assert
        SecurityContext.User actingUser = securityContext.getActingUser().get();
        SecurityContext.User.Credentials credentials = actingUser.getCredentials().get();
        assertEquals( Optional.of("1234"), credentials.getEducationCode());
        assertEquals(Optional.empty(), credentials.getNationalRole());
        assertEquals("1234", serviceActor.getUserTitle());
    }

    @Test
    public void getServiceActor_forHealthCareProfessionalWithEducationCodeTitleIsMapped() {
        // Act
        mockActingHealthcareProfessionalUser("5265", null);
        ServiceActor serviceActor = serviceActorProvider.getServiceActor(securityContext, false);

        // Assert
        SecurityContext.User actingUser = securityContext.getActingUser().get();
        SecurityContext.User.Credentials credentials = actingUser.getCredentials().get();
        assertEquals( Optional.of("5265"), credentials.getEducationCode());
        assertEquals(Optional.empty(), credentials.getNationalRole());
        assertEquals("Kiropraktor", serviceActor.getUserTitle());
    }

    @Test
    public void getServiceActor_forHealthCareProfessionalWithNationalRole() {
        // Act
        mockActingHealthcareProfessionalUser(null, "1234");;
        ServiceActor serviceActor = serviceActorProvider.getServiceActor(securityContext, false);

        // Assert
        SecurityContext.User actingUser = securityContext.getActingUser().get();
        SecurityContext.User.Credentials credentials = actingUser.getCredentials().get();
        assertEquals( Optional.empty(), credentials.getEducationCode());
        assertEquals(Optional.of("1234"), credentials.getNationalRole());
        assertEquals("1234", serviceActor.getUserTitle());
    }

    @Test
    public void getServiceActor_forHealthCareProfessionalWithNationalRoleTitleIsMapped() {
        // Act
        mockActingHealthcareProfessionalUser(null, "urn:dk:healthcare:national-federation-role:code:41002:value:SundAssistR2");;
        ServiceActor serviceActor = serviceActorProvider.getServiceActor(securityContext, false);

        // Assert
        SecurityContext.User actingUser = securityContext.getActingUser().get();
        SecurityContext.User.Credentials credentials = actingUser.getCredentials().get();
        assertEquals( Optional.empty(), credentials.getEducationCode());
        assertEquals(Optional.of("urn:dk:healthcare:national-federation-role:code:41002:value:SundAssistR2"), credentials.getNationalRole());
        assertEquals("Sundhedsdata-Læser", serviceActor.getUserTitle());
    }

    @Test
    public void getServiceActor_forHealthCareProfessionalWithNoCredentials() {
        // Act
        mockActingHealthcareProfessionalUser(null, null);
        ServiceActor serviceActor = serviceActorProvider.getServiceActor(securityContext, false);

        // Assert
        SecurityContext.User actingUser = securityContext.getActingUser().get();
        SecurityContext.User.Credentials credentials = actingUser.getCredentials().get();
        assertEquals(Optional.empty(), credentials.getEducationCode());
        assertEquals(Optional.empty(), credentials.getNationalRole());
        assertEquals(UNKNOWN_USER_TITLE, serviceActor.getUserTitle());
    }

    @Test
    public void getServiceActor_unknownOrganisationIdFormat() {
        // Arrange
        SecurityContext.Organisation organisation = Mockito.mock(SecurityContext.Organisation.class);
        Mockito.when(organisation.getIdentifierFormat()).thenReturn(null);
        Mockito.when(securityContext.getOrganisation()).thenReturn(Optional.of(organisation));

        // Act
        try {
            serviceActorProvider.getServiceActor(securityContext, false);
            fail();
        } catch (Exception e) {
            // Assert
            assertEquals(NullPointerException.class, e.getClass());
        }
    }

    @Test
    public void getServiceActor_unknownPersonIdFormat() {
        // Arrange
        SecurityContext.User actingUser = Mockito.mock(SecurityContext.User.class);
        Mockito.when(actingUser.getUserType()).thenReturn(SecurityContext.User.UserType.Citizen);
        Mockito.when(actingUser.getIdentifierFormat()).thenReturn(null);
        Mockito.when(securityContext.getActingUser()).thenReturn(Optional.of(actingUser));

        // Act
        try {
            serviceActorProvider.getServiceActor(securityContext, false);
            fail();
        } catch (Exception e) {
            // Assert
            assertEquals(NullPointerException.class, e.getClass());
        }
    }

    @Test
    public void getServiceActor_forHealthCareProfessionalWithPrincipalUser_fails() {
        // Arrange
        SecurityContext.User actingUser = Mockito.mock(SecurityContext.User.class);
        Mockito.when(actingUser.getUserType()).thenReturn(SecurityContext.User.UserType.HealthcareProfessional);
        Mockito.when(securityContext.getPrincipalUser()).thenReturn(Optional.of(Mockito.mock(SecurityContext.User.class)));

        // Act
        try {
            serviceActorProvider.getServiceActor(securityContext, false);
            fail();
        } catch (Exception e) {
            // Assert
            assertEquals(FaultCode.AUTHORIZATION_ERROR_INGEN_DELEGERING.toString(), e.getMessage());
        }
    }

    @Test
    public void getServiceActor_forHealthCareProfessionalAdmin() {
        // Act
        ServiceActor serviceActor = serviceActorProvider.getServiceActor(securityContext, true);

        // Assert
        assertEquals(UserType.Admin, serviceActor.getUserType());
        assertEquals("7170", serviceActor.getUserTitle());
    }

    @Test
    public void getServiceActor_forHealthCareProfessionalWithoutAudience() {
        // Arrange
        SecurityContext.Ticket ticketWithoutAudience = createTicketWithoutAudience();
        Mockito.when(securityContext.getTicket()).thenReturn(Optional.of(ticketWithoutAudience));

        // Act
        ServiceActor serviceActor = serviceActorProvider.getServiceActor(securityContext, false); // Succeeds
        assertEquals("7170", serviceActor.getUserTitle());
    }

    @Test
    public void getServiceActor_forHealthCareProfessionalWithInvalidAudience() {
        // Arrange
        serviceActorProvider = new ServiceActorProvider("http://gyldig", whitelistedCvrs, userTitleProvider);

        // Act
        ServiceActor serviceActor = serviceActorProvider.getServiceActor(securityContext, false); // Succeeds
        assertEquals("7170", serviceActor.getUserTitle());
    }

    @Test
    public void getServiceActor_forHealthCareProfessionalNotWhitelisted() {
        // Arrange
        serviceActorProvider = new ServiceActorProvider(audience, Collections.emptyList(), userTitleProvider); // Given cvr no longer whitelisted

        // Act
        ServiceActor serviceActor = serviceActorProvider.getServiceActor(securityContext, false); // Succeeds
        assertEquals("7170", serviceActor.getUserTitle());
    }

    @Test
    public void getServiceActor_forCitizen() {
        // Arrange
        mockActingUser(SecurityContext.User.UserType.Citizen);

        // Act
        ServiceActor serviceActor = serviceActorProvider.getServiceActor(securityContext, false);

        // Assert
        assertEquals(UserType.Citizen, serviceActor.getUserType());
        assertEquals(BORGER_USER_TITLE, serviceActor.getUserTitle());
    }

    @Test
    public void getServiceActor_forCitizenNotWhitelisted() {
        // Arrange
        mockActingUser(SecurityContext.User.UserType.Citizen);
        serviceActorProvider = new ServiceActorProvider(audience, Collections.emptyList(), userTitleProvider); // Given cvr no longer whitelisted

        // Act
        ServiceActor serviceActor = serviceActorProvider.getServiceActor(securityContext, false); // succeeds

        assertEquals(BORGER_USER_TITLE, serviceActor.getUserTitle());
    }

    @Test
    public void getServiceActor_forCitizenAdmin_fails() {
        // Arrange
        mockActingUser(SecurityContext.User.UserType.Citizen);

        // Act
        try {
            serviceActorProvider.getServiceActor(securityContext, true);
            fail();
        } catch (Exception e) {
            // Assert
            assertEquals(FaultCode.AUTHORIZATION_ERROR_USERTYPE_ADMIN_NOT_ALLOWED.toString(), e.getMessage());
        }
    }

    @Test
    public void getServiceActor_forCitizenWithoutAudience_fails() {
        // Arrange
        mockActingUser(SecurityContext.User.UserType.Citizen);
        SecurityContext.Ticket ticketWithoutAudience = createTicketWithoutAudience();
        Mockito.when(securityContext.getTicket()).thenReturn(Optional.of(ticketWithoutAudience));

        // Act
        try {
            serviceActorProvider.getServiceActor(securityContext, false);
            fail();
        } catch (Exception e) {
            // Assert
            assertEquals(FaultCode.AUTHORIZATION_ERROR_INVALID_HEADER.toString(), e.getMessage());
        }
    }

    @Test
    public void getServiceActor_forCitizenWithInvalidAudience_fails() {
        // Arrange
        mockActingUser(SecurityContext.User.UserType.Citizen);
        serviceActorProvider = new ServiceActorProvider("http://gyldig", whitelistedCvrs, userTitleProvider);

        // Act
        try {
            serviceActorProvider.getServiceActor(securityContext, false);
            fail();
        } catch (Exception e) {
            // Assert
            assertEquals(FaultCode.AUTHORIZATION_ERROR_WRONG_AUDIENCE.toString(), e.getMessage());
        }
    }

    @Test
    public void getServiceActor_forCitizenWithPrincipalUser_fails() {
        // Arrange
        mockActingUser(SecurityContext.User.UserType.Citizen);
        Mockito.when(securityContext.getPrincipalUser()).thenReturn(Optional.of(Mockito.mock(SecurityContext.User.class)));

        // Act
        try {
            serviceActorProvider.getServiceActor(securityContext, false);
            fail();
        } catch (Exception e) {
            // Assert
            assertEquals(FaultCode.AUTHORIZATION_ERROR_INGEN_DELEGERING.toString(), e.getMessage());
        }
    }

    @Test
    public void getServiceActor_forSystemUserAdmin_fails() {
        // Arrange
        // No acting user, but an organization, means system user
        Mockito.when(securityContext.getActingUser()).thenReturn(Optional.empty());

        // Act
        try {
            serviceActorProvider.getServiceActor(securityContext, true);
            fail();
        } catch (Exception e) {
            // Assert
            assertEquals(FaultCode.AUTHORIZATION_ERROR_USERTYPE_ADMIN_NOT_ALLOWED.toString(), e.getMessage());
        }
    }

    @Test
    public void getServiceActor_forSystemUserNotWhitelisted_fails() {
        // Arrange
        // No acting user, but an organization, means system user
        Mockito.when(securityContext.getActingUser()).thenReturn(Optional.empty());
        serviceActorProvider = new ServiceActorProvider(audience, Collections.emptyList(), userTitleProvider); // Given cvr no longer whitelisted

        // Act
        try {
            serviceActorProvider.getServiceActor(securityContext, false);
            fail();
        } catch (Exception e) {
            // Assert
            assertEquals(FaultCode.AUTHORIZATION_ERROR_CVR_NOT_WHITELISTED.toString(), e.getMessage());
        }
    }

    @Test
    public void getServiceActor_forSystemUserWithInvalidAudience() {
        // Arrange
        // No acting user, but an organization, means system user
        Mockito.when(securityContext.getActingUser()).thenReturn(Optional.empty());
        serviceActorProvider = new ServiceActorProvider("http://gyldig", whitelistedCvrs, userTitleProvider);

        // Act
        ServiceActor serviceActor = serviceActorProvider.getServiceActor(securityContext, false); // succeeds
        assertNull(serviceActor.getUserTitle());
    }

    @Test
    public void getServiceActor_forSystemWithPrincipalUser() {
        // Arrange
        // No acting user, but an organization, means system user
        Mockito.when(securityContext.getActingUser()).thenReturn(Optional.empty());
        Mockito.when(securityContext.getPrincipalUser()).thenReturn(Optional.of(Mockito.mock(SecurityContext.User.class)));

        // Act
        ServiceActor serviceActor = serviceActorProvider.getServiceActor(securityContext, false);
        assertNull(serviceActor.getUserTitle());
    }

    private void mockActingUser(SecurityContext.User.UserType userType) {
        SecurityContext.User actingUser = Mockito.mock(SecurityContext.User.class);
        Mockito.when(actingUser.getUserType()).thenReturn(userType);
        Mockito.when(actingUser.getIdentifierFormat()).thenReturn(SecurityContext.User.PersonIdentifierFormat.CPR);
        Mockito.when(securityContext.getActingUser()).thenReturn(Optional.of(actingUser));
    }

    private SecurityContext.Ticket createTicketWithoutAudience() {
        SecurityContext.Ticket ticket = Mockito.mock(SecurityContext.Ticket.class);
        Mockito.when(ticket.isValid()).thenReturn(true);
        Mockito.when(ticket.getAudience()).thenReturn(Optional.empty());
        return ticket;
    }

    private void mockActingHealthcareProfessionalUser(String educationCode, String nationalRole) {
        SecurityContext.User actingUserMock = Mockito.mock(SecurityContext.User.class);
        SecurityContext.User.Credentials credentials = Mockito.mock(SecurityContext.User.Credentials.class);
        Mockito.when(actingUserMock.getUserType()).thenReturn(SecurityContext.User.UserType.HealthcareProfessional);
        Mockito.when(actingUserMock.getIdentifierFormat()).thenReturn(SecurityContext.User.PersonIdentifierFormat.CPR);
        Mockito.when(actingUserMock.getCredentials()).thenReturn(Optional.of(credentials));
        Mockito.when(credentials.getEducationCode()).thenReturn(Optional.ofNullable(educationCode));
        Mockito.when(credentials.getNationalRole()).thenReturn(Optional.ofNullable(nationalRole));
        Mockito.when(securityContext.getActingUser()).thenReturn(Optional.of(actingUserMock));
        ServiceActor serviceActor = serviceActorProvider.getServiceActor(securityContext, false);
    }
}
